from typing import Any

from strawberry.types.info import Info


class BasePermission:
    """
    Base class for creating permissions
    """

    def has_permission(self, source: Any, info: Info, **kwargs) -> bool:
        raise NotImplementedError(
            "Permission classes should override has_permission method"
        )



from itertools import chain
from typing import Optional, cast

from graphql.type import is_object_type, is_specified_directive
from graphql.utilities.print_schema import (
    is_defined_type,
    print_args,
    print_block,
    print_deprecated,
    print_description,
    print_directive,
    print_implemented_interfaces,
    print_schema_definition,
    print_type as original_print_type,
)

from strawberry.types.types import FieldDefinition, TypeDefinition

from .schema import BaseSchema


def print_federation_field_directive(field: Optional[FieldDefinition]) -> str:
    if not field:
        return ""

    out = ""

    if field.federation.provides:
        out += f' @provides(fields: "{field.federation.provides}")'

    if field.federation.requires:
        out += f' @requires(fields: "{field.federation.requires}")'

    if field.federation.external:
        out += " @external"

    return out


def print_fields(type_, schema: BaseSchema) -> str:
    strawberry_type = cast(TypeDefinition, schema.get_type_by_name(type_.name))

    fields = []

    for i, (name, field) in enumerate(type_.fields.items()):
        field_definition = strawberry_type.get_field(name) if strawberry_type else None

        fields.append(
            print_description(field, "  ", not i)
            + f"  {name}"
            + print_args(field.args, "  ")
            + f": {field.type}"
            + print_federation_field_directive(field_definition)
            + print_deprecated(field)
        )

    return print_block(fields)


def print_federation_key_directive(type_, schema: BaseSchema):
    strawberry_type = cast(TypeDefinition, schema.get_type_by_name(type_.name))

    if not strawberry_type:
        return ""

    keys = strawberry_type.federation.keys

    parts = []

    for key in keys:
        parts.append(f'@key(fields: "{key}")')

    if not parts:
        return ""

    return " " + " ".join(parts)


def print_extends(type_, schema: BaseSchema):
    strawberry_type = cast(TypeDefinition, schema.get_type_by_name(type_.name))

    if strawberry_type and strawberry_type.federation.extend:
        return "extend "

    return ""


def _print_object(type_, schema: BaseSchema) -> str:
    return (
        print_description(type_)
        + print_extends(type_, schema)
        + f"type {type_.name}"
        + print_federation_key_directive(type_, schema)
        + print_implemented_interfaces(type_)
        + print_fields(type_, schema)
    )


def _print_type(field, schema: BaseSchema) -> str:
    if is_object_type(field):
        return _print_object(field, schema)

    return original_print_type(field)


def print_schema(schema: BaseSchema) -> str:
    graphql_core_schema = schema._schema  # type: ignore

    directives = filter(
        lambda n: not is_specified_directive(n), graphql_core_schema.directives
    )
    type_map = graphql_core_schema.type_map

    types = filter(is_defined_type, map(type_map.get, sorted(type_map)))  # type: ignore

    return "\n\n".join(
        chain(
            filter(None, [print_schema_definition(graphql_core_schema)]),
            (print_directive(directive) for directive in directives),
            (_print_type(type_, schema) for type_ in types),  # type: ignore
        )
    )



from decimal import Decimal
from typing import Optional
from uuid import UUID

import pydantic

from .exceptions import UnsupportedTypeError


FIELDS_MAP = {
    pydantic.NoneStr: Optional[str],
    pydantic.NoneBytes: Optional[bytes],
    pydantic.StrBytes: None,
    pydantic.NoneStrBytes: None,
    pydantic.StrictStr: str,
    pydantic.ConstrainedBytes: bytes,
    pydantic.conbytes: bytes,
    pydantic.ConstrainedList: None,
    pydantic.conlist: None,
    pydantic.ConstrainedSet: None,
    pydantic.conset: None,
    pydantic.ConstrainedStr: str,
    pydantic.constr: str,
    pydantic.EmailStr: str,
    pydantic.PyObject: None,
    pydantic.ConstrainedInt: int,
    pydantic.conint: int,
    pydantic.PositiveInt: int,
    pydantic.NegativeInt: int,
    pydantic.ConstrainedFloat: float,
    pydantic.confloat: float,
    pydantic.PositiveFloat: float,
    pydantic.NegativeFloat: float,
    pydantic.ConstrainedDecimal: Decimal,
    pydantic.condecimal: Decimal,
    pydantic.UUID1: UUID,
    pydantic.UUID3: UUID,
    pydantic.UUID4: UUID,
    pydantic.UUID5: UUID,
    pydantic.FilePath: None,
    pydantic.DirectoryPath: None,
    pydantic.Json: None,
    pydantic.JsonWrapper: None,
    pydantic.SecretStr: str,
    pydantic.SecretBytes: bytes,
    pydantic.StrictBool: bool,
    pydantic.StrictInt: int,
    pydantic.StrictFloat: float,
    pydantic.PaymentCardNumber: None,
    pydantic.ByteSize: None,
    pydantic.AnyUrl: str,
    pydantic.AnyHttpUrl: str,
    pydantic.HttpUrl: str,
    pydantic.PostgresDsn: str,
    pydantic.RedisDsn: str,
}


def get_basic_type(type_):
    if isinstance(type_, type):
        if issubclass(type_, pydantic.ConstrainedInt):
            return int
        if issubclass(type_, pydantic.ConstrainedStr):
            return str

    if type_ in FIELDS_MAP:
        type_ = FIELDS_MAP.get(type_)

        if type_ is None:
            raise UnsupportedTypeError()

    return type_



from typing import cast

from strawberry.scalars import is_scalar
from strawberry.types.types import FieldDefinition


def _convert_from_pydantic_to_strawberry_field(
    field: FieldDefinition, data_from_model=None, extra=None
):
    data = data_from_model or extra

    if field.is_list:
        child = field.child
        items = [None for _ in data]

        for index, item in enumerate(data):
            items[index] = _convert_from_pydantic_to_strawberry_field(
                cast(FieldDefinition, child),
                data_from_model=item,
                extra=extra[index] if extra else None,
            )

        return items
    elif is_scalar(field.type):  # type: ignore
        return data
    else:
        return convert_pydantic_model_to_strawberry_class(
            field.type, model_instance=data_from_model, extra=extra
        )


def convert_pydantic_model_to_strawberry_class(cls, *, model_instance=None, extra=None):
    extra = extra or {}
    kwargs = {}

    for field in cls._type_definition.fields:
        field = cast(FieldDefinition, field)
        origin_name = field.origin_name

        data_from_extra = extra.get(origin_name, None)
        data_from_model = (
            getattr(model_instance, origin_name, None) if model_instance else None
        )
        kwargs[origin_name] = _convert_from_pydantic_to_strawberry_field(
            field, data_from_model, extra=data_from_extra
        )

    return cls(**kwargs)
